/****************************************************************
   File Edge.h
   Defines and implements the connection between two milestones
   on the tree. It is composed by information on both nodes
   (their indexes), the distance between them and the status on 
   collision checking for the straight-line segment that joins 
   both milestones on the C-space.
   Kappa is an integer >=0 which is used to compute how many points
   inside a segment have been already checked. For a segment with
   kappa equal to k, it means that it has been checked for
   2^(k) equally-spaced points.  

   Gildardo Sanchez-Ante
   Robotics Lab.,  Stanford University
*****************************************************************/

#ifndef _EDGE_H_
#define _EDGE_H_
#include <math.h>
class Edge{
  
 public:
 
  Edge(){ 
    from = 0;
    to = 0;
    lambda = 1000;
    kappa = 0;
  };
  
  Edge(int f, int t){
    from = f;
    to = t;
    lambda = 1000;
    kappa = 0;
  };

  Edge(int f, int t, double d, int k){
    from = f;
    to = t;
    lambda = d;
    kappa = k;
  };
  
  void setFrom(int f){from =f;}
  void setTo(int t){to= t;}
  void setDist(double d){lambda = d;}
  void setKappa(int k){kappa=k;}
  void markSafe(){kappa = -1;}
  void markBad(){kappa = -100;}
  void setEdge(int f, int t, double d=0, int k=0){
    from =f;  to = t;  lambda =d; kappa = k;}

  int getFrom(){return from;};
  int getTo(){return to;};
  double getDist(){return lambda;};
  int getKappa(){return kappa;}; 

  bool isSafe(){
    if(kappa == -1) return true;
    else return false;
  };

  void print(){
    cerr<<"Link from "<<from<<" to  "<<to<<
      " distance: "<<lambda<<" K: "<<kappa<<'\n';
  };

  friend class PrioritizeEdges;

 protected:
  int        from;  // Reference of the node where the link starts
  int        to;    // Reference of the node where the link ends
  double     lambda;  // distance between the two milestones
  int        kappa; // how many times we have partitioned the segment
};

// Friend class with function object for comparison of Objects
class PrioritizeEdges {
  public :
    int operator()( const Edge x, const Edge y )
    {
      return pow(2,-x.kappa)*x.lambda < pow(2,-y.kappa)*y.lambda;
    }
};

#endif
