#ifndef GEN_COORDS_H
#define GEN_COORDS_H

#include <iostream.h>
#include <stdio.h>
#include <math.h>

const int MAX_SIZE=12;			


/* GenCoords
 *    generalized coordinates for the configuration of a robot
 */

class GenCoords  {

 public:
  double x[MAX_SIZE];

  GenCoords();
  GenCoords(const double xx[]);
  GenCoords(const float xx[]);

  GenCoords(const GenCoords& u);

  GenCoords& operator=(const GenCoords& u);

  const double& operator[] (int i) const;
  double& operator [] (int i);

  double Norm() const;			// Euclidean norm
  double Norm2() const;			// Euclnorm  norm ^ 2
  double Dist(const GenCoords& y) const;
  double Max() const;			// L-infinity norm

  static void Dim(int d);
  static int Dim();
  void print()
    { 
      printf("%f  %f  %f  %f  %f  %f\n",
	     x[0],x[1],x[2],x[3],x[4],x[5]);
    }
  
private:
  static int dim_;
};

istream& operator>> (istream& is, GenCoords& q);
ostream& operator<< (ostream& os, const GenCoords& q);


inline
GenCoords::GenCoords()
{
}


inline 
GenCoords::GenCoords(const double xx[])
{
    memcpy(x, xx, dim_*sizeof(double));
}

inline 
GenCoords::GenCoords(const float xx[])
{
  for(int i=0;i<dim_;i++)
    x[i]=double(xx[i]);
}
inline 
GenCoords::GenCoords(const GenCoords& u)
{
    memcpy(x, u.x, dim_*sizeof(double));
}

inline GenCoords&
GenCoords::operator=(const GenCoords& u)
{
    memcpy(x, u.x, dim_*sizeof(double));
    return *this;
}


inline const double&
GenCoords::operator[] (int i) const
{
    return x[i];
}

inline double&
GenCoords::operator[] (int i) 
{
    return x[i];
}


inline double
GenCoords::Norm2() const
{
    int i=dim_;
    double l=0.0;

    while (i--)  l += x[i] * x[i];
    return l;
}

inline double
GenCoords::Dist(const GenCoords& y) const
{
    int i=dim_;
    double l=0.0;

    while (i--)  l += (x[i]-y.x[i]) * (x[i]-y.x[i]);
    return sqrt(l);
}

inline double
GenCoords::Max() const
{
    int i=dim_;
    double l=0.0;
    double y;

    while (i--) {
      y = x[i] > 0.0 ? x[i] : -x[i];
      l = y > l ? y : l;
    }

    return l;
}

inline double
GenCoords::Norm() const
{
    int i=dim_;
    double l=0.0;

    while (i--)  l += x[i] * x[i];
    return sqrt(l);
}

inline int
GenCoords::Dim()
{
  return dim_;
}

inline void
GenCoords::Dim(int d)
{
  dim_ = d;
}


/* operator + 
 *    add two vectors 
 */

inline GenCoords
operator+ (const GenCoords& u, const GenCoords& v)
{
    int i=GenCoords::Dim();
    GenCoords w;

    while (i--) w.x[i] = u.x[i] + v.x[i];
    return w; 
}

/* operator -
 *    subtract two vectors 
 */

inline GenCoords
operator- (const GenCoords& u, const GenCoords& v)
{
    int i=GenCoords::Dim();
    GenCoords w;

    while (i--) w.x[i] = u.x[i] - v.x[i];
    return w; 
}

/* operator *
 *    multiply by a scalar
 */

inline GenCoords
operator* (double a, const GenCoords& v)
{
    int i=GenCoords::Dim();
    GenCoords w;

    while (i--) w.x[i] = a * v.x[i];
    return w; 
}

/* operator /
 *    divide by a scalar
 */

inline GenCoords
operator/ (const GenCoords& v, double a)
{
  return 1.0 / a * v;
}


/* operator /
 *    component-wise division
 */

inline GenCoords
operator/ (const GenCoords& u, const GenCoords& v)
{
    int i=GenCoords::Dim();
    GenCoords w;

    while (i--) w.x[i] = u.x[i] / v.x[i];
    return w; 
}


/* unary operator-
 *    return the nagated vector
 */

inline  GenCoords
operator- (const GenCoords& v)
{
  int i=GenCoords::Dim();
  GenCoords w;

  while (i--) w.x[i] = -v.x[i];
  return w; 
}


/* dot
 *    inner product of two vectors 
 */

inline double
dot(const GenCoords& u, const GenCoords& v)
{
    int i=GenCoords::Dim();
    double l=0.0;

    while (i--) l += u.x[i] * v.x[i];
    return l; 
}


#endif
