// **************************************************************
// SBL Path Planner Declaration  
// One robot in a static environment, bidirectional search
// lazy-collision checking
// 
// Gildardo Sanchez-Ante 
// Robotics Lab. Stanford University
// ante@robotics.stanford.edu
// *************************************************************

#ifndef PLANNER_H
#define PLANNER_H

#include <sys/times.h>
#include <stdlib.h>
#include <queue>
#include <vector>

#include "Milestone.h"
#include "Path.h"
#include "pumaivsimulator.h"
#include "Rn.h"
#include "myTree.h"
#include "Edge.h"

typedef priority_queue<Edge,vector<Edge>, PrioritizeEdges> edgesQueue;

enum {IN_SAME_BIN, ANY_BIN, ANY_NODE, NEAREST_NODE};
enum {T_INIT=0, T_GOAL};

class Planner {
  
 public:
  
  Planner(IVSimulator *s);
  virtual ~Planner();

  bool Query( const GenCoords& q0, const GenCoords& q1, 
	      Path& path, double rho=0.15, int s=10000);

  void printStats(struct tms start, struct tms end, Path& );
  void writePath(list<Milestone>, char *);
  
 protected:

  // Functions to build the tree
  void expandTree(double rho, Milestone& m);
  bool ConnectTrees(int i, const Milestone m);
  bool sample(const Milestone& q, Milestone& q_rand, double);


  // Functions to update tree connectivity
  void delSegment(Edge& inColl);
  void updateBins(int tree, Milestone& m);
  void delMilestone(int tree, Milestone& m);

  // For Path
  bool testPath(edgesQueue&);
  bool testSegment(Edge& u);

  // For the management of the edges
  int getIndex(const Milestone& x0, const Milestone& x1);
  bool setRoot(const GenCoords& q, myTree& T);
  void addMilestone(myTree& T, Milestone& to_add, Milestone father);
  void getEdges(Milestone&, Milestone&, edgesQueue& );
  void getPath(list<Milestone>& p);

  // Optimization of the path
  // void optimize();
  bool optimizePath();
  void trimPath(int from, int to);

  // For extracting statistics
  void checkPath(const list<Milestone> p);
  void report_times (struct tms start, struct tms end);
  void addTime(struct tms start, struct tms end);

  static bool initRand; //random number initializer flag
  double EPSILON; // resolution static bool initRand; 
  int tree_; // ID for the trees
  int nodes_counter;
  int forbid;
  struct tms optstart, optend; // to measure time

  // The two trees, one rooted at the initial config and the other
  // rooted at the goal config
  myTree Tg;
  myTree Ti;
  int dof0_;
  int dof1_;

  // the set of all the nodes generated on both trees
  Milestone allNodes[10000];

  Edge theBridge; // To store the bridge

  list<Milestone> freePath; //To store the first path found
  list<Milestone> optimalPath;// To store the optimized path

  IVSimulator *ivsim; // Pointer to the simulator

  Rn randConfig;

  // Variables used for keeping info related to the performance
  int onBridge;
  int path_rej;
  int sampled_Mil;
  int total_CC;   // Counter of total coll checks 
  int path_CC;    // Counter of coll checks when checking the path
  double CC_time; // To store the time required for Collision-Checking
  int dofCounter[6][10];
};

#endif 
