//////////////////////////////////////////////////////////////
//
// Header file for SimMainWindow
//
//    This class is a subclass of VkWindow
//
// Normally, very little in this file should need to be changed.
// Create/add/modify menus using RapidApp.
//
// Restrict changes to those sections between
// the "//--- Start/End editable code block" markers
// Doing so will allow you to make changes using RapidApp
// without losing any changes you may have made manually
//
//
// Modified by Gildardo Sanchez-Ante
// Robotics Lab, Stanford University
//////////////////////////////////////////////////////////////


#ifndef SIMMAINWINDOW_H
#define SIMMAINWINDOW_H
#include <Vk/VkWindow.h>


class VkMenuItem;
class VkMenuToggle;
class VkMenuConfirmFirstAction;
class VkSubMenu;
class VkRadioSubMenu;

//---- Start editable code block: headers and declarations
class Milestone;
class EditMainWindow;
class ViewMainWindow;
class IVSimulator;
class SoSeparator;
class SoSwitch;
class Sample;

#include "Path.h"
#include "SampleRandomConfig.h"
#include "Rn.h"
#include "Hpaned.h"

//---- End editable code block: headers and declarations


//---- SimMainWindow class declaration

class SimMainWindow: public VkWindow {

  public:

    SimMainWindow( const char * name, IVSimulator *sim_ = NULL);
    ~SimMainWindow();
    const char *className();
    virtual Boolean okToQuit();

    //---- Start editable code block: SimMainWindow public
     
    class SoXtExaminerViewer *viewer();

    bool openRobotFile(const char *filename);
    bool openWorkspaceFile(const char *filename);
    IVSimulator *getSimulator();
    void setSimulator(IVSimulator *sim_);
    void Draw(const Path& p,int n);
    void Draw(const Path& p1,const Path& p2);

    //---- End editable code block: SimMainWindow public


  protected:

    // Classes created by this class

    class Hpaned *_hpaned;

    // Widgets created by this class

    // Menu items created by this class
    VkSubMenu  *_filePane;
    VkMenuItem *_openRobotButton;
    VkMenuItem *_openWorkspaceButton;
    VkMenuItem *_separator1;
    VkMenuItem *_exitButton;

    VkSubMenu    *_viewPane;
    VkMenuToggle *_editModeToggle;
    VkMenuToggle *_viewModeToggle;
    VkMenuToggle *_noModeToggle;

    VkSubMenu  *_demosPane;
    
    VkMenuItem *_setDemo1Button;
    VkMenuItem *_setDemo2Button;
    VkMenuItem *_setDemo3Button;
    VkMenuItem *_setDemo4Button;
    VkMenuItem *_setDemo5Button;
    VkMenuItem *_setDemo6Button; 
    VkMenuItem *_setDemo7Button;
  
    VkSubMenu  *_operationsPane;
    VkMenuItem *_setRobotsButton;
    VkMenuItem *_separator2;
    VkMenuItem *_colCheckButton;
    VkMenuItem *_distCompButton;
    VkMenuItem *_separator3;
    VkMenuItem *_setInitConfButton;
    VkMenuItem *_setGoalConfButton;
    VkMenuItem *_separator4;
    VkMenuItem *_runPlannerButton;
    VkMenuItem *_animatePathButton;


    // Member functions called from callbacks

    // File Menu
    virtual void doOpenRobot ( Widget, XtPointer );
    virtual void doOpenWorkspace ( Widget, XtPointer );
    virtual void doQuit ( Widget, XtPointer );

    // Edit Menu
    virtual void doSetEditMode ( Widget, XtPointer );
    virtual void doSetNoMode ( Widget, XtPointer );
    virtual void doSetViewMode ( Widget, XtPointer );

    // Demos Menu
    
    virtual void doSetDemo1 ( Widget, XtPointer );
    virtual void doSetDemo2 ( Widget, XtPointer );
    virtual void doSetDemo3 ( Widget, XtPointer );
    virtual void doSetDemo4 ( Widget, XtPointer );
    virtual void doSetDemo5 ( Widget, XtPointer );
    virtual void doSetDemo6 ( Widget, XtPointer ); 
    virtual void doSetDemo7 ( Widget, XtPointer );


      // Operations Menu
    virtual void doSetRobots ( Widget, XtPointer );
    virtual void doColCheck ( Widget, XtPointer );
    virtual void doDistComp ( Widget, XtPointer );   
    virtual void doSetInitConf ( Widget, XtPointer );
    virtual void doSetGoalConf ( Widget, XtPointer );
    virtual void doRunPlanner ( Widget, XtPointer );
    virtual void doAnimatePath( Widget, XtPointer );
    //---- Start editable code block: SimMainWindow protected


    //---- End editable code block: SimMainWindow protected 

  private:
    // Callbacks to interface with Motif

    static void openWorkspaceCallback ( Widget, XtPointer, XtPointer );
    static void openRobotCallback ( Widget, XtPointer, XtPointer );
    static void quitCallback ( Widget, XtPointer, XtPointer );


    static void setEditModeCallback ( Widget, XtPointer, XtPointer );
    static void setNoModeCallback ( Widget, XtPointer, XtPointer );
    static void setViewModeCallback ( Widget, XtPointer, XtPointer );

    
    static void setDemo1Callback ( Widget, XtPointer, XtPointer );
    static void setDemo2Callback ( Widget, XtPointer, XtPointer );
    static void setDemo3Callback ( Widget, XtPointer, XtPointer );
    static void setDemo4Callback ( Widget, XtPointer, XtPointer );
    static void setDemo5Callback ( Widget, XtPointer, XtPointer );
    static void setDemo6Callback ( Widget, XtPointer, XtPointer ); 
    static void setDemo7Callback ( Widget, XtPointer, XtPointer );

    static void setRobotsCallback ( Widget, XtPointer, XtPointer );
    static void colCheckCallback ( Widget, XtPointer, XtPointer );
    static void distCompCallback ( Widget, XtPointer, XtPointer );
    
    static void setInitConfCallback ( Widget, XtPointer, XtPointer );
    static void setGoalConfCallback ( Widget, XtPointer, XtPointer );
    static void runPlannerCallback ( Widget, XtPointer, XtPointer );
    static void animatePathCallback(Widget, XtPointer, XtPointer );

    static String  _defaultSimMainWindowResources[];


    //---- Start editable code block: SimMainWindow private

  enum {mrkFrame, mrkBall};

  EditMainWindow *editWindow[10];
  ViewMainWindow *viewWindow;

  IVSimulator *sim;
  SoSeparator *root;
 
  Path freePath[10];
  Path optimalPath[10];
  Path::iterator iterOnPath[10]; 
  
  SoSeparator *getDecorations();
  SoSeparator *createDecor();
  SoSeparator *createFloor();
  SoSeparator *createSky();
  SoSwitch* getWristPath(const Path& p, const Path& ip);

  SoSwitch* getWristPath(const Path& p, const Path& ip, float scale, int r =0);
  SoSeparator* createMarker(int mtype, float scale);
  SoSeparator* createBall(float scale);
  SoSeparator* createFrame(float scale);
  SoSeparator* createFloor(float size, int density);

  GenCoords q_goal; 
  GenCoords q_init;

  //---- End editable code block: SimMainWindow private


};
//---- Start editable code block: End of generated code


//---- End editable code block: End of generated code

#endif
















