// ***************************************************
//   File ivsimulator.h
//   Declaration of class IVSimulator.
//   It defines all the environment for the simulator
//   i.e., robots, obstacles, floor, etc.
//   >>>>>>>>>>>>>   PQP VERSION   <<<<<<<<<<<<<<<<<<
//
//   Modified by Gildardo Sanchez-Ante
//   Robotics Lab Stanford University
// ****************************************************

#ifndef ivsimulator_h
#define ivsimulator_h

#include "RobotSimulator.h"
#include "ObstacleKit.h"
#include "Obstacle.h"
#include "GenCoords.h"

#include <Inventor/SoLists.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoCallbackAction.h>


// Maximum number of objects on the workspace
#define MAX_OBJS  10
#define MAX_ROBS 10
#define ROBOTS 1

class SoGroup;
class SoSeparator;
class Milestone;

class IVSimulator : public RobotSimulator {
public:
  IVSimulator(SoSeparator *root_ = NULL);
  virtual ~IVSimulator();
 
  GenCoords maxVelocity;	  // normalized max velocity for each dof
  GenCoords jointInterval;

  static int objs;                // Objects on the environment

  // Inherited

  virtual bool IsClear(const Milestone& m, int rob=0);

  virtual bool IsClear(const Milestone& m0, const Milestone& m1, 
		       IntervalType itype=Closed);
  virtual bool IsClear(const Path& p);
  
  bool areSafe(SoPathList& jks0,const Milestone& m0,
	       SoPathList& jks1,const Milestone& m1);

  bool areSafe(const Milestone& m0, const Milestone& m1);

  // IVSimulator specific
  virtual void GetConfig(GenCoords *q, int rob =0); 
  virtual void SetConfig(const GenCoords& q, int rob=0);
  
  virtual float getCurDist(int rob =0);
  void printConfig(const GenCoords& q);
  int getNumJoints() {return jointKitsR[0].getLength();}
  const char *getJointKitName(int jindex);

  virtual void setRoot(SoSeparator *root_);
  SoSeparator *getRobot(int rob=0) {return robotSep[rob];}
  SoSeparator *getRoot() {return root;}
  
  bool readRobot(const char *infile);
  bool readWorkspace(const char *infile);

  static void initInventor();

 protected:

  // Methods
 
  void initialize();
  void normalizeConfig(const GenCoords& q, GenCoords& q0);
  void denormalizeConfig(const GenCoords& q0, GenCoords& q);
  static void findJointKits(SoNode *top, SoPathList& jointKits);
  static void findObstacleKits(SoNode *top, SoPathList& jointKits);
  void decomposeObjKit(SoPathList jointKit);

  // To modify the pose of the robot referenced by the SoPathList
  static void setConfig(SoPathList& jks, const GenCoords& q);
  // To modify the position of the obstacle given by obstID
  void setLocation(int obstID, double x, double y, double z); 
 
  SoGroup *getSeg(const GenCoords& q0, const GenCoords& q1,
		  IntervalType itype=SemiClosed);
  SoGroup *getPath(const Path& p);

  // Data Members

  float vmaxR; //Maximum velocity for the robot

  // Pointers to construct the scene graph
  SoSeparator *root; // The root of the scene graph
  SoSeparator *robot[MAX_ROBS];
  SoSeparator *robotSep[MAX_ROBS]; // one separator for each robot  
  SoSeparator *backRobot[2]; // a copy of the shape of the robot, used
                             // for collision checking purposes
  
  SoSeparator *theEnvironment; // Dummy variable. Used to store the
                               // environment read from an iv file
  
  SoPathList jointKitsR[MAX_ROBS];
  SoPathList obstKits;
  SoPathList brJointKits[2];

  // Each object on the environment needs pointers to
  // define its shape (obstacleSep), and its 
  // position (xformObst), as well as a collision-checking
  // model (colModObst)
  // shapeModObst is a dummy variable needed to extract the
  // objects from the iv files
  SoSeparator    *obstacleSep[MAX_OBJS]; 
  SoTransform    *xformObst[MAX_OBJS];  
  PQP_Model      *colModObst[MAX_OBJS];
  Obstacle       *shapeModObst[MAX_OBJS];

};

#endif










