
//////////////////////////////////////////////////////////////
//
// Source file for EditMainWindow
//
//    This class is a subclass of VkSimpleWindow
//
//
// Normally, very little in this file should need to be changed.
// Create/add/modify menus using RapidApp.
//
// Try to restrict any changes to the bodies of functions
// corresponding to menu items, the constructor and destructor.
//
// Restrict changes to those sections between
// the "//--- Start/End editable code block" markers
//
// Doing so will allow you to make changes using RapidApp
// without losing any changes you may have made manually
//
// Avoid gratuitous reformatting and other changes that might
// make it difficult to integrate changes made using RapidApp
//
// Modified by Gildardo Sanchez-Ante
// Robotics Lab, Stanford University
//
//////////////////////////////////////////////////////////////

#include "EditMainWindow.h"

#include <Vk/VkApp.h>
#include <Vk/VkResource.h>
#include "ivsimulator.h"

// Externally defined classes referenced by this class: 

#include "EditBoard.h"

extern void VkUnimplemented ( Widget, const char * );

//---- Start editable code block: headers and declarations

#include <iostream.h>
#include <Vk/VkEZ.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoSelection.h>
#include <Inventor/manips/SoTransformBoxManip.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include "jointkit.h"
#include "joint.h"
#include "link.h"

int EditMainWindow::rob = 0;

//---- End editable code block: headers and declarations



// These are default resources for widgets in objects of this class
// All resources will be prepended by *<name> at instantiation,
// where <name> is the name of the specific instance, as well as the
// name of the baseWidget. These are only defaults, and may be overriden
// in a resource file by providing a more specific resource name

String  EditMainWindow::_defaultEditMainWindowResources[] = {

        //---- Start editable code block: EditMainWindow Default Resources


        //---- End editable code block: EditMainWindow Default Resources

        (char*)NULL
};


//---- Class declaration

EditMainWindow::EditMainWindow ( const char *name,
                                ArgList args,
                                Cardinal argCount) : 
                          VkSimpleWindow ( name, args, argCount )
{
    // Load any class-default resources for this object

    setDefaultResources ( baseWidget(), _defaultEditMainWindowResources  );


    // Create the view component contained by this window

    _editBoard = new EditBoard ( "editBoard",mainWindowWidget() );
    

    XtVaSetValues ( _editBoard->baseWidget(),
                    XmNwidth, 337, 
                    XmNheight, 323, 
                    (XtPointer) NULL );

    // Add the component as the main view

    addView ( _editBoard );

    // Set up the window manager properties for this window.
    // This window is a coprimary window, as defined by the SGI Style Guide
    // This window cannot be resized by the user.
    // This window has no title bar.

     VkConfigureWM ( this,
                    FALSE,   //quit
                    TRUE,  // close
                    TRUE,   // border
                    FALSE,  // title
                    FALSE,  // allow resize
                    FALSE,  // allow iconify
                    FALSE);  // Show window manager menu

    //---- Start editable code block: EditMainWindow constructor
	
	sim = NULL;
	_editBoard->setEditor(this);
	selJointKit = NULL;
	manipRoot = new SoSelection;
	manipNode = new SoTransformBoxManip;
	manipRoot->addChild(manipNode);

	jointViewer()->setSceneGraph(manipRoot);
	
	//manipNode->getDragger()->addMotionCallback(modifyManip, &selJointKit);
	
    //---- End editable code block: EditMainWindow constructor


}    // End Constructor


EditMainWindow::~EditMainWindow()
{
    delete _editBoard;
    //---- Start editable code block: EditMainWindow destructor


    //---- End editable code block: EditMainWindow destructor
}    // End destructor


const char *EditMainWindow::className()
{
    return ("EditMainWindow");
}    // End className()


Boolean EditMainWindow::okToQuit()
{
    //---- Start editable code block: EditMainWindow okToQuit



    // This member function is called when the user quits by calling
    // theApplication->terminate() or uses the window manager close protocol
    // This function can abort the operation by returning FALSE, or do some.
    // cleanup before returning TRUE. The actual decision is normally passed on
    // to the view object


    // Query the view object, and give it a chance to cleanup

    return ( _editBoard->okToQuit() );

    //---- End editable code block: EditMainWindow okToQuit
}    // End okToQuit()





/////////////////////////////////////////////////////////////// 
// The following functions are called from callbacks 
/////////////////////////////////// 




//---- Start editable code block: End of generated code

void EditMainWindow::setSim(IVSimulator *sim_) {
	sim = sim_;
	resetSim();
}

void EditMainWindow::resetSim() {
	XmString *slItems = new XmString[sim->getNumJoints()];
	for (int i = 0; i < sim->getNumJoints(); i++) {
		char *name = new char[strlen(sim->getJointKitName(i))+1];
		strcpy(name, sim->getJointKitName(i));
		slItems[i] = XmStringCreateLocalized(name);
	}
	
	XtVaSetValues ( *jointList(), XmNitems, slItems,
			XmNitemCount, sim->getNumJoints(), NULL );
}	

void EditMainWindow::setCamera(SoCamera *cam) {
	jointViewer()->getCamera()->position.connectFrom(&cam->position);
	jointViewer()->getCamera()->orientation.connectFrom(&cam->orientation);			
	jointViewer()->getCamera()->focalDistance.connectFrom(&cam->focalDistance);
	cam->position.connectFrom(&jointViewer()->getCamera()->position);
	cam->orientation.connectFrom(&jointViewer()->getCamera()->orientation);
	cam->focalDistance.connectFrom(&jointViewer()->getCamera()->focalDistance);
}

//This is used for one of the robots
void EditMainWindow::selectJointKit(const char *jointKitName) {
 
  // if (selJoint)
  //selJoint->disconnectInvEngine();
 
  if (sim != NULL) {
    SoSearchAction searchAction;
    searchAction.setName(jointKitName);
    searchAction.setInterest(SoSearchAction::FIRST);
    searchAction.apply(sim->getRobot(num_rob));
 
    /*
      fprintf(stderr,"NROB %d  ",num_rob); 
      switch(num_rob)
      {
      case 0:
      searchAction.apply(sim->getRobot(0));
      break;
      default:
      searchAction.apply(sim->getRobot(1));
      break;
      }
    */
    
    SoFullPath *path = (SoFullPath *)searchAction.getPath();
    
    if (path != NULL) {
      selJointKit = (JointKit *)path->getNodeFromTail(0);
      selJoint = (Joint *)selJointKit->getPart("joint", FALSE);
    }
    
    if (selJoint != NULL) {
      int newVal = selJoint->getValAt(0)*100.;
      XtVaSetValues(*_editBoard->dofScale(), XmNvalue, 
		    newVal, NULL);
      XtVaSetValues(*_editBoard->dofNum(), XmNvalue, 0, NULL);
      XtVaSetValues(*_editBoard->dofNum(), XmNmaximum, 
		    selJoint->getNumVals(), NULL);
      selJoint->connectInvEngine(manipNode);
    }
  }
}

void EditMainWindow::setDof(int dof, double newVal) {
	if (selJoint != NULL) {
		selJoint->setValAt(dof, newVal);
	}		

	//resetModel(selJointKit);
}


Widget *EditMainWindow::jointList() {
	return _editBoard->jointList();
}

SoXtExaminerViewer *EditMainWindow::jointViewer() {
	return _editBoard->jointViewer();
}

/*
void EditMainWindow::modifyManip(void *userData, SoDragger *dragger) {
	resetModel(*((JointKit **)userData));
}


void EditMainWindow::resetModel(JointKit *kit) {
	if (kit != NULL) {
		kit->updateMatrix(true);
	}
}
*/

//---- End editable code block: End of generated code











