// **************************************************************
// SBL Path Planner implementation.
// One robot in a static environment, bidirectional search
// lazy-collision checking
//
//  Gildardo Sanchez-Ante
//  Robotics Lab.    Stanford University
//  ante@Robotics.Stanford.edu
// **************************************************************

#include <stdlib.h>
#include <sys/times.h>
#include <stdio.h>

#include "Milestone.h"
#include "Planner.h" 
#include "Path.h"
#include "myTree.h"

// Class variable initializations
bool Planner::initRand = true;

// ------------------------------
// Constructor & Destructor
// ------------------------------
Planner::Planner(IVSimulator *sim) 
{ 
  EPSILON = 0.012; // Epsilon is the resolution to which a segment is
                   // checked in order to say that it is collision-free
                   // this value corresponds at approx. 200 CC on the 
                   // longest line in C space (sqrt(6)/200)

  ivsim = sim;
  nodes_counter =0;
  forbid=0;
 
  if (initRand){
    srand48(time(NULL)); // Seed the random number generator.
    initRand = false; // Only once though :-)
  } 
  dof0_ = 0;
  dof1_ = 1;

  Ti.setDOFS(dof0_,dof1_);
  Tg.setDOFS(dof0_,dof1_);

  // Initialize the trees
  Ti.setGridSize(10,10); // Grid size for both trees
  Tg.setGridSize(10,10);
  
  total_CC=0;     
  path_CC=0;
  path_rej=0;     
  sampled_Mil=0;
  CC_time=0;      
  onBridge=0;
}

// Destructor
Planner::~Planner() 
{}



// ----------------------------------------------------------------
// Query: receive the start and goal configurations (qs and qg) and
//        returns the path (genPath) if found. rho = neighborhood to
//        sample, s = max number of iterations allowed
// ----------------------------------------------------------------
bool
Planner::Query(const GenCoords& qs, const GenCoords& qg, 
               Path& genPath, double rho, int s)
{
  struct tms start,end; // to take time for planning 

  if(! setRoot(qs,Ti)){
    cerr << "Initial configuration in collision!"<<endl;
    return false;
  }

  if(! setRoot(qg,Tg)){
    cerr << "Goal configuration in collision!"<<endl;
    return false;
  }

  int iter = 2;

  times(&start);

  // Algorithm PLANNER
  while(iter<s){
    Milestone new_node;

    expandTree(rho, new_node);
    if (ConnectTrees(IN_SAME_BIN, new_node))
      break;
    if((iter % 50)==0){
      myTree tmpTi;
      myTree tmpTg;
      do{
	dof0_ = (lrand48() % 6);
	dof1_ = (lrand48() % 6);
      }while(dof0_==dof1_);
      
      tmpTi.setDOFS(dof0_,dof1_);
      tmpTg.setDOFS(dof0_,dof1_);
      
      // Initialize the trees
      tmpTi.setGridSize(10,10); 
      tmpTg.setGridSize(10,10);
      
      // Copy the milestones
      Ti.rebuildTree(tmpTi, allNodes);
      Tg.rebuildTree(tmpTg, allNodes);
      Ti = tmpTi;
      Tg = tmpTg;
    }
    iter++;
  };
  times(&end);
  
  if(iter>s){
    cerr << "Sorry, couldn't found a path" << endl;
    return false;
  }
  else {
    printStats(start, end, genPath);
    //optimize();
    return true;
  }
  
}

// ----------------------------------------------------------------
// expandTree: it is the responsible of growing the trees
//             Basically, it selects a node from the tree 
//             to expand, and then samples in its neighborhood 
//             then add the new node to the tree             
// ----------------------------------------------------------------
void 
Planner::expandTree(double rho, Milestone& q_new)
{
  double rand_n = drand48();
  if(rand_n>=0.5) // Expand each tree with prob = 0.5
    { // Expand Tg
      // Choose a node from the tree to be expanded
      int selected = Tg.getNodeToExpand();
      // Get the node from the tree
      Milestone q_expand = allNodes[selected];

      // Get q_new by sampling on an rho-neighborhood of q_expand
      while(! sample(q_expand, q_new, rho)){
	// sample couldn't found a free config, try again with
	// a different node to expand
	selected = Tg.getNodeToExpand(); 
	q_expand = allNodes[selected];
      };
 
      // Add the q_new to the tree
      addMilestone(Tg, q_new,q_expand);
      tree_ = T_GOAL; // to signal that the expanded tree was Tg
    }
  else // Expand Ti
    {
      // Same process than in the if part...
      int selected = Ti.getNodeToExpand(); 
      Milestone q_expand = allNodes[selected];

      while(! sample(q_expand, q_new, rho)){
	selected = Ti.getNodeToExpand();
	q_expand = allNodes[selected];
      };
 
      addMilestone(Ti, q_new, q_expand);
      tree_ = T_INIT;
    }
  return;
}
 
// ---------------------------------------------------------
// sample: It receives a milestone m and generate one new
//         milestone in the rho-neighborhood of m
// ---------------------------------------------------------
bool
Planner::sample(const Milestone& m, Milestone& new_node, double rho)
{
  bool isInCollision;
  GenCoords conf;
  int i=1;  double rho_local= rho;
  
  do{
    randConfig.LocalUniformSample(m.getConfig(),rho_local,&conf);
    new_node.setConfig(conf);
    total_CC++;
    sampled_Mil++;
 
    times(&optstart);
    isInCollision = !(ivsim->IsClear(new_node));
    times(&optend);
    addTime(optstart, optend);

    if(! isInCollision)
      return true;
    
    // If the sampled config was in collision, reduce the size of rho
    rho_local = rho/(i);
    i++;
  }while(i<4);
  // it couldn't found a free config after 4 iters, then quit
  return false;
}

// ----------------------------------------------------------------
// Function ConnectTrees:   It tries to establish connection
//                          between two nodes, one from each
//                          tree.
// ----------------------------------------------------------------
bool
Planner::ConnectTrees(int option,  Milestone m)
{ 
  int selNode;
  Milestone a, b;
    
  // ConnectTrees try a connection between the new node and any other
  // node in the same bin on the other tree

  // Compute in which bin is the new node
  GenCoords q = m.getConfig();
  double delta = Ti.getDelta();
  int cells = Ti.getDIVS(); 

  int selBin = int(floor(q[dof0_]/delta)+floor(q[dof1_]/delta)*cells);
  
  // Get a node on the same bin on the other tree
  if(tree_ == T_INIT)// if the expanded tree was Ti
    selNode = Tg.getClosestNode(m, allNodes)/*Tg.getNodeFromBin(selBin)*/;
  
  if(tree_ == T_GOAL)
    selNode = Ti.getClosestNode(m, allNodes)/*Ti.getNodeFromBin(selBin)*/;

  // If no node was selected, connectTrees has failed
  if(selNode == -1){
    if(tree_ == T_INIT){// if the expanded tree was Ti
      int bin = Tg.chooseBin();
      selNode = Tg.getNodeFromBin(bin);
      double d = allNodes[m.getIndex()].distance(allNodes[selNode]);
      if(d >0.8)
	selNode = -1;
    }
    if(tree_ == T_GOAL){
      int bin = Ti.chooseBin();
      selNode = Ti.getNodeFromBin(bin);
      double d = allNodes[m.getIndex()].distance(allNodes[selNode]);
      if(d >0.8)
	selNode = -1;
    }
  }
  
  if(selNode == -1)
    return false;
  
  // Otherwise, get the indexes for the two nodes that were
  // connected

  if(tree_ == T_INIT){
    a = allNodes[m.getIndex()];
    b = allNodes[selNode];
  }
  if(tree_ == T_GOAL){// Must be T_GOAL
    a = allNodes[selNode];
    b = allNodes[m.getIndex()]; 
  }

  // Constructs the bridge between the two trees
  theBridge.setFrom(a.getIndex());
  theBridge.setTo(b.getIndex());
  theBridge.setEdge(a.getIndex(),b.getIndex(),a.distance(b));
 
  // Put the path in a Priority Queue sorted according to the 
  // length of the segments in decreasing order
  edgesQueue candidatePath;
  getEdges(a, b, candidatePath);

  return (testPath(candidatePath));
}


// -----------------------------------------------------------
// testPath():  Tests whether a path is clear or not at 
//              a certain resolution. It uses the partition
//              technique
// -----------------------------------------------------------
bool 
Planner::testPath(edgesQueue& tau)
{
  while(!tau.empty()){
    // Take the segment which is on the top of the queue
    // i.e. the one with longest distance between nodes
    Edge u =  tau.top();
    tau.pop();

    // Test the segment
    if(testSegment(u) == false){
      delSegment(u);
      return false;
    }
    // If segment wasn't in collision (at this resolution) but it 
    // cannot be marked as safe yet, re insert it on the queue
    if(! u.isSafe())
      tau.push(u);
  }
  return true;
}


// -------------------------------------------------------------------
// testSegment:Given the former kappa (k) of a segment, this function
//             checks the segment a kappa=k+1 to see if it is clear.
//             If the segment is free at a resolution smaller than 
//             epsilon, the segment is marked as safe (-1).  
// -------------------------------------------------------------------
bool 
Planner::testSegment(Edge& u)
{
  double lambda = u.getDist();
  int k = u.getKappa();

  if(u.isSafe())
    return true;
 
  Milestone& m0 = allNodes[u.getFrom()];
  Milestone& m1 = allNodes[u.getTo()];
  Milestone ml,mr;

  // Computes the max number of steps required for that kappa
  int max_n=1;
  int i;
  for(i=1;i<k+1;i++)
    max_n*=2;
  //int max_n = int(pow(2,k+1)); // cambiar por un ciclo!

  // Computes the distance to be increased on each coordinate
  GenCoords delta_q = (m1.getConfig()-m0.getConfig())/max_n;
  
  // If we have a segment |0----1----2----3----4----5----6----7----8|
  // and we were on Kappa = 3, max_n = 8, then, we will start
  // checking from node 4 towards the initial and final points on the
  // segment.
  GenCoords q0 = m0.getConfig() + (max_n/2)*delta_q;
  bool isOK;//  int j=1;
  for( i=0;i<(max_n/4);i++){
    // From midpoint to the left
    ml.setConfig(q0 - (2*i+1)*delta_q);
    
    total_CC++; 
    times(&optstart);
       isOK = ivsim->IsClear(ml);
    times(&optend);
    addTime(optstart, optend);
    
    if(! isOK)
      return false;
    
    // From the midpoint to the right
    mr.setConfig(q0 + (2*i+1)*delta_q);

    total_CC++; 
    times(&optstart);
       isOK = ivsim->IsClear(mr);
    times(&optend);
    addTime(optstart, optend);
    
    if(! isOK)
      return false;
  }
  
  if((lambda/max_n) < EPSILON){
    u.markSafe();
    return true;
  }
  else{
    u.setKappa(k+1); // This modifies the copy received
    allNodes[u.getTo()].setKappa(k+1);// This modifies the actual value
    return true;
  }
}

// -------------------------------------------------------
// setRoot: used to set  the root of the tree
// -------------------------------------------------------
bool
Planner::setRoot(const GenCoords& q, myTree& T)
{    
  Milestone m(nodes_counter, q);

  total_CC++; // Update the counter of coll checks
  if ( ! ivsim->IsClear(m) ) { 
    cerr << "Configuration in collision!"<<endl;
    return false;
  }

  allNodes[nodes_counter] = m;
  T.addToBins(m);
  ++nodes_counter;
  return true;
}


// ----------------------------------------------------------------
// addMilestone: Adds a milestone to the tree, and updates the bins 
// ----------------------------------------------------------------
void 
Planner:: addMilestone(myTree& T, Milestone& newNode, Milestone expandedNode)
{
  newNode.setIndex(nodes_counter);
  int x = expandedNode.getIndex();

  // The edge on allNodes[x] is modified
  allNodes[x].isFatherOf(newNode);
  allNodes[nodes_counter] =  newNode;
  T.addToBins(newNode);       
  ++nodes_counter;

  // This code is to compute how the nodes are diffusing
  GenCoords cfg = newNode.getConfig();
  for(int i=0;i<6;i++)    
      dofCounter[i][int(floor(cfg[i]*10))]++;
}

// ---------------------------------------------------------------------
// getEdges(): Once that all the segments have been marked as clear,
// this function extracts all the edges in such a way that creates the
// actual path
// ---------------------------------------------------------------------
void 
Planner::getEdges(Milestone &node1, Milestone& node2, edgesQueue& pathQueue)
{
  freePath.clear();

  int from =  node1.getFather();
  Edge tmpEdge = node1.getEdge();
  
  if(from==-999){
    freePath.push_back(node1);
  }
  else {
    freePath.push_back(node1);
    //Extracts the segments on Ti
    do{    
      pathQueue.push(tmpEdge);
      Milestone next = allNodes[from];

      freePath.push_front(next);
      tmpEdge = next.getEdge(); 
      from = next.getFather();

    } while(from!=-999);
  }

  from = node2.getFather();
  tmpEdge =  node2.getEdge();
 
  if(from==-999){
    freePath.push_back(node2);
  }
  else {
    //Extracts the segments on Tg
    freePath.push_back(node2);
    do{    
      pathQueue.push(tmpEdge);
      Milestone next = allNodes[from];
      
      freePath.push_back(next);
      tmpEdge = next.getEdge();
      from = next.getFather();
    } while(from !=-999);
  }

  pathQueue.push(theBridge);

  return;
}

// ---------------------------------------------------------------
// delSegment: When a collision is detected in one of the
//             edges, eliminate the edge from the tree
// ---------------------------------------------------------------
void 
Planner::delSegment(Edge& inColl)
{
  int inTree= 0;
  
  // Get the milestones of the edge to be eliminated
  int from = inColl.getFrom();
  int to = inColl.getTo();
   
  // If they coincide with the bridge, nothing to be done here
  if(from  == theBridge.getFrom()){
    return;
  }
  
  // Take node 'from' and go backwards until the root is reached
  // This is for finding out in which tree is the bad link 
  Edge tmp = allNodes[from].getEdge();
  
  while(tmp.getFrom()!=-999){
    Milestone next = allNodes[tmp.getFrom()];
    tmp = next.getEdge();
  };

  // Each tree has as root a node with 'from' = -999 and 'to'= x
  // where x is an integer different for each tree. Ti = 0, Tg = 1 ...
  inTree = tmp.getTo(); 

  int a, b, c;
  Edge theFatherOfb;

  if(inTree == T_INIT){ // The broken link is on Ti
    a = theBridge.getTo();
    b = theBridge.getFrom();
    theFatherOfb = allNodes[b].getEdge();
    c = theFatherOfb.getFrom();
  }
  else{ // It is on Tg
    a = theBridge.getFrom();
    b = theBridge.getTo();
    theFatherOfb = allNodes[b].getEdge();
    c = theFatherOfb.getFrom();
  }
  
  //This establishes the direction of the connection on the bridge
  allNodes[a].isFatherOf(allNodes[b]); // a is the father of b
  allNodes[b].setKappa(theBridge.getKappa());

  updateBins(inTree, allNodes[b]);
  
  while(b != to ){ 
    a = b;
    b = c;
    theFatherOfb = allNodes[b].getEdge();
    c = theFatherOfb.getFrom();
    
    allNodes[a].isFatherOf(allNodes[b]); // a is father of b
    allNodes[b].removeChild(allNodes[a]); // a is not any more a child of b
    
    updateBins(inTree, allNodes[b]);
  }
  allNodes[c].removeChild(allNodes[b]); // b is not any more a child of c
}

// -----------------------------------------------------------------
// updateBins: The purpose of this function is to update the bins of
//             trees Ti and Tg, deleting the milestone m from one of
//             them and adding it to the other one 
// -----------------------------------------------------------------
void 
Planner::updateBins(int del, Milestone& m)
{
  int sons = m.getNrSons();
  if(sons == 0){ // If the milestone doesn't have children
    if(del==T_INIT){ 
      // Delete the node from one bin and add it to the other
      Ti.delFromBins(m);
      Tg.addToBins(m);
    }
    else{ //Delete from T_GOAL
      Tg.delFromBins(m);
      Ti.addToBins(m);
    }
    
    return; // end with the recursion
  }

  else{ 
    // Otherwise, there are children. Do the same for
    // each one of them.
    int r;
    for(int k=0;k<sons;k++){
      r = m.getChild(k);
      updateBins(del, allNodes[r]);
    }
    // Once that all the children have been deleted,
    // delete also the root of this subtree
    delMilestone(del,m);
  }
}

/* --------------------------------------------------------------------
   This function is used as an auxiliary one to delete one node from 
   the tree, not taking care of the child. 
   -------------------------------------------------------------------- 
*/
void 
Planner::delMilestone(int del, Milestone& m)
{
  if(del==T_INIT){ 
    // Delete the node from one bin and add it to the other
    Ti.delFromBins(m);
    Tg.addToBins(m);
  }
  else{ //Delete from T_GOAL
    Tg.delFromBins(m);
    Ti.addToBins(m);
  }
  return;
}

/* --------------------------------------------------------------------
   Function optimize Path: Used to shorten the path trying to avoid some
   unnecessary movements
   -------------------------------------------------------------------- 
*/
bool Planner::optimizePath()
{ 
  int nodesOnPath = optimalPath.size();

  // If there are too few milestones on the path, do not even try to
  // shorten it
  if ( nodesOnPath<=4)
    return false;

  int a, b;
  
  do{
    // get two random nodes on the path
    a = lrand48() % nodesOnPath;
    b = lrand48() % nodesOnPath;
  }while (a==b || (abs(a-b)==1) ); // if the nodes are equal,     

  int from, to;
  // to be sure that the smaller number is the 'from'
  if(a > b){
    from = b;
    to = a;
  } 
  else {
    from = a;
    to = b;
  } 
   
  // Extract the milestones given by 'from' and 'to' 
  Milestone ma, mb;
  list<Milestone>::iterator I;
  int cnt = 0;
  
  for(I = optimalPath.begin();I!=optimalPath.end();I++){
    if(cnt == from)
      ma = (*I);
    if(cnt == to){
      mb = (*I);
      break;
    }
    cnt++;
  }

  if(ivsim->IsClear(ma,mb)){
    // If the straight-line path between ma and mb is clear
    trimPath(from,to); // trim the path between those nodes
    return true;

  }
  else{
    // Computes its midpoint and recursively do the same process
    // int midP = ((to-from)/2) + from;
    // optimizePath(from, midP);
    // optimizePath(midP,to);
    return false;
  }
}


/* -----------------------------------------------------------------
   Function trimPath: Used to trim a segment of the path. 
                      The segment was already checked to be free
   -----------------------------------------------------------------
*/
void Planner::trimPath(int from, int to)
{
  list<Milestone> Tau;
  // Copy the milestones from 0 to 'from' to Tau
  // Tau is a buffer to store the trimmed path
  list<Milestone>::iterator I = optimalPath.begin();
  int cnt = 0;
  for(cnt =0; cnt<=from; cnt++){
    Tau.push_back(*I);
    I++;
  }
  // Skip milestones between 'from' and 'to'
  for(; cnt<to; cnt++)
    I++;

  // Copy milestones from 'to' to the end of the path on Tau
  for(; I!=optimalPath.end(); cnt++){
    Tau.push_back(*I);
    I++;
  }

  list<Milestone>::iterator I2;
  int  nodes = 0;
  optimalPath.clear();
  for(I2 = Tau.begin();I2!=Tau.end();I2++){
    (*I2).setIndex(nodes++);
    optimalPath.push_back(*I2);
  } 
}

// **************************************************************
//                  Auxiliary I/O Functions
// **************************************************************

// --------------------------------------------------------------
// report_times: Print a report of time consumption
// --------------------------------------------------------------
void 
Planner::report_times (struct tms start, struct tms end)
{
  float ticks = CLK_TCK;
  float utime = (end.tms_utime - start.tms_utime) / ticks;
  float stime = (end.tms_stime - start.tms_stime) / ticks;
  // cerr << "Total time: " << utime + stime << "\n";
  cerr << "\n"<<utime + stime << "\t";    
 }

// --------------------------------------------------------------
// printStats: Presents statistics of the trees and path
// --------------------------------------------------------------
void 
Planner::printStats(struct tms start, struct tms end, Path& genPath)
{
  report_times(start, end);
  checkPath(freePath);
  // This prints a line with:
  // time|Nr Mil|Mil on Path|Total CC|CC Path|sampled Mil|

  cerr<<Ti.getSize()+Tg.getSize()<<"\t"<<freePath.size()<<"\t"\
      <<total_CC<<"\t"<<path_CC<<"\t"<<sampled_Mil<<"\t"<<CC_time; 
  
  //checkPath(freePath);
  writePath(freePath, "freePath.txt");
  
  // Copy the contents of freePath to optimalPath, and
  // renumerates the indexes of the milestones on it from 0, 1,...
  int cnt =0;
  list<Milestone>::iterator I;
  for(I = freePath.begin();I!=freePath.end();I++){
    (*I).setIndex(cnt);
    optimalPath.push_back(*I);
    cnt++;
  }
  
  // Run the optimization process until 5 reductions on the size
  // of the path are achieved or we have tried 10 times without getting
  // any further improvement on the path
  int reductions =0, trials = 0;
  //cerr<<"Optimizing the path ..... please wait"<<endl;
  
  while ((reductions < 15) && (trials < 15)){
    if(optimizePath()==true)
      reductions++;
    else 
      trials++;
  }
  
  checkPath(optimalPath);

  // Copy the path to the parameter sent to the function
  for(I = optimalPath.begin();I!=optimalPath.end();I++) 
    genPath.push_back(*I);
  
  writePath(optimalPath, "optimalPath.txt");
}

// ------------------------------------------------------------------
// AddTime: To store the time required for collision checking
// ------------------------------------------------------------------
void 
Planner::addTime(struct tms start, struct tms end){
  float ticks = CLK_TCK;
  float utime = (end.tms_utime - start.tms_utime) / ticks;
  float stime = (end.tms_stime - start.tms_stime) / ticks;
  CC_time+= (utime + stime); // time
}

// --------------------------------------------------------------
// writePath: Sends the path to a file
// --------------------------------------------------------------
void Planner::
writePath(list<Milestone> path, char* name)
{
  list<Milestone>::const_iterator mi;
  Milestone v;
  GenCoords q;
  FILE *fp;

  fp=fopen(name,"wt");

  if (fp==NULL){
    cerr<<"Error in Planner::write_path when opening file"<<endl;
    return ;
  }
  
  for ( mi = path.begin();
        mi != path.end();
        mi++)
    { 
      v = (*mi);
      q=v.getConfig();
      fprintf(fp, "%f  %f  %f  %f  %f  %f\n",
              q[0],q[1],q[2],q[3],q[4],q[5]);
    }
  fclose(fp);
}

/* ------------------------------------------------------------------
   Function pathStats: This function re-checks all the path in order
   to get statistics on the minimum number of collison- check
   operations required for the planner.
   ------------------------------------------------------------------
*/
void Planner::checkPath(const list<Milestone> toCheck)
{
  int cc = 0; // Counter of collision checks
  Milestone current, next;
  GenCoords delta_q;
  double dm=0.0, dist=0.0; // distance traveled by end effector in Wspace
  int steps=0, pathSize =0;

  pathSize = toCheck.size();
  list<Milestone>::const_iterator pathIter = toCheck.begin();

  current = (*pathIter);

  for( int counter =0;counter <pathSize-1;counter++){
    next = (*(++pathIter));
    dist += current.distance(next);
    delta_q = (next.getConfig()-current.getConfig());
    dm = delta_q.Max();
    steps = int(ceil(dm/EPSILON));
    
    Milestone m;
    for(int i=0;i <= steps;i++){
      path_CC++;
      cc++;
      m.setConfig(current.getConfig() + i*delta_q);
      ivsim->IsClear(m);
    }
    current = next;
  };
}
