/*  **********************************************************************
    File ivsimulator.C
    Implements class IVSimulator
    
    This is may be the most important file from the whole
    hierarchy, since it is here where the complete environment
    for the simulator is defined. 

    >>>>>>>>>>>>>>>>>>>>>>>   PQP VERSION    <<<<<<<<<<<<<<<<<<<<<<<<<<<<<
 

    Modified by Gildardo Sanchez-Ante
    Robotics Lab, Stanford University
    ante@Robotics.stanford.edu
    **************************************************************************/

#include <iostream.h>
#include "ivsimulator.h"
#include "ivgeneral.h"  // .h files for the IVSimulator

#include <PQP.h>

#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoShape.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/SoXtRenderArea.h>

// Setting variable dim to the number of DOF of the robot
int GenCoords::dim_ = 6;

// initializes the counter of objects
int IVSimulator::objs = 0;

// ----------------------------------------------------------------
// Constructor
// ----------------------------------------------------------------
IVSimulator::IVSimulator(SoSeparator *root_) {
  
  objs=0; // Counter of obstacles
  if (!root_) {
    root_ = new SoSeparator;
    root_->ref();
  }
  
  root = root_; // The root of the scene graph
  
  // Creates and references one separator for each robot
  for (int j=0;j<ROBOTS;j++){  
    robotSep[j] = new SoSeparator; 
    robotSep[j] -> ref();
    root->addChild(robotSep[j]);
    robot[j]=NULL;
  }
  
  // For the objects on the environment
  for (int k=0;k<MAX_OBJS;k++){
    shapeModObst[k] = NULL; // The shape (graphics) 
    colModObst[k]= NULL;    // The collision-checking model 
  }
  backRobot[0] = NULL; 
  backRobot[1] = NULL;
}

// ------------------------------------------------------------
// Destructor
// ------------------------------------------------------------
IVSimulator::~IVSimulator() {}

// ------------------------------------------------------------------
// readRobot: For reading the iv file with the definition of the robot
// -------------------------------------------------------------------
bool 
IVSimulator::readRobot(const char *infile) {
  SoInput sceneInput;
  int numDofs = 0;
  int i,j, k;

  // For all the robots, open and read the iv file with the definition
  // and add the shape to the scene graph
  for(i=0;i<ROBOTS;i++){
    if(robot[i]){
      robotSep[i]->removeChild(robot[i]);
      robot[i]->unref();
    }
    sceneInput.openFile(infile);
    robot[i] = SoDB::readAll(&sceneInput);    
  }
  
  for(j =0; j<ROBOTS;j++){ // two back robots
    if (backRobot[j])
      backRobot[j]->unref();
    
    sceneInput.openFile(infile);
    backRobot[j] = SoDB::readAll(&sceneInput);
  }
  
  // Decompose the joint kits to get the links
  for (k=0;k<ROBOTS;k++){
  
    robot[k]->ref();
    robotSep[k]->addChild(robot[k]);
    
    findJointKits(robotSep[k], jointKitsR[k]);
    
    for (i = 0; i < jointKitsR[k].getLength(); i++) {
      JointKit *jk = (JointKit *)(((SoNodeKitPath *)jointKitsR[k][i])->
				  getNodeFromTail(0));
      Joint *joint = (Joint *)jk->getPart("joint", FALSE);
      Link *link = (Link *)jk->getPart("link", FALSE);
      if (joint) {
	numDofs += joint->getNumVals();
      }
      // construct the collision model for each link
      if (link)
	link->buildColModel(jk);
    }
  } // end for the robots
  
  for(k=0;k<ROBOTS;k++){
    backRobot[k]->ref();
    findJointKits(backRobot[k], brJointKits[k]);
  
    for (int j = 0; j < brJointKits[k].getLength(); j++) {
      JointKit *jk = (JointKit *)(((SoNodeKitPath *)brJointKits[k][j])\
				  ->getNodeFromTail(0));
      Joint *joint = (Joint *)jk->getPart("joint", FALSE);
      Link *link=(Link *)jk->getPart("link", FALSE);
      if (link)
	link->buildColModel(jk);
    } 
  } // end for the back robots

  initialize();
  
  return true;
}

// ------------------------------------------------------------------
// readObject: To read the iv file with the definition of the obstacle
// ------------------------------------------------------------------
bool 
IVSimulator::readWorkspace(const char *infile) {
  // This function reads an iv file in which an environment is
  // defined. The environment may include several objects, and each
  // one of them should be managed independently from the rest
  SoInput    sceneInput;
  
  if ( !sceneInput.openFile(infile) )
    return false;
 
  // Read the file with the environment (this can be composed of several
  // objects
  if (theEnvironment) {     
    root->removeChild(theEnvironment);
    theEnvironment->unref();
  }                   
                
  theEnvironment = SoDB::readAll(&sceneInput);

  if (!theEnvironment)
    return false;

  // Given that the file with the environment can be complex
  // (in the sense that it may have several objects, at different
  // positions, the following two functions are responsible to
  // extract each independent object (its shape)
  findObstacleKits(theEnvironment, obstKits);
  decomposeObjKit(obstKits);

  return true;
}

// -----------------------------------------------------------------
// IsClear: Checks if the robot rob at config m is in collision or not
//          basically, it takes each of the links of the robot and
//          checks them againts all the objects on the environment
// -----------------------------------------------------------------
bool 
IVSimulator::IsClear(const Milestone& m, int rob) 
{
  GenCoords q(m.getConfig());

  setConfig(brJointKits[rob], q);

  // This part of the code checks the links of each robot
  // for collision against all the objects on the environment
  int i, j=0;
  for (j=0;j<objs;j++){ // And for each obstacle
    for(i = 0; i < brJointKits[rob].getLength(); i++) {
      
      // Extract each one of the links of the robot
      JointKit *jk = (JointKit *)(((SoNodeKitPath *) brJointKits[rob][i])
				  ->getNodeFromTail(0));
      Link *link = (Link *)jk->getPart("link", FALSE);
      
      // And see whether the link is in collision with the obstacle or not
      if (link && (!link->isClear(shapeModObst[j])))
	return false;
    }
  }
  return true;
}

// --------------------------------------------------------------------
// IsClear: Used to check if a straigh-line path from m0 to m1 is clear
// --------------------------------------------------------------------
bool 
IVSimulator::IsClear(const Milestone& m0, const Milestone& m1, 
			  IntervalType itype) {
  cerr<< "ERROR: IsClear::IVSimulator Not implemented!"<<endl;
}

// ------------------------------------------------------------
// IsClear: To check a sequence of milestones
// ------------------------------------------------------------
bool 
IVSimulator::IsClear(const Path& p) {
  cerr<< "ERROR: IsClear::IVSimulator Not implemented!"<<endl;
}


// ------------------------------------------------------------
// areSafe: This function checks if robot jks0 and jks1 at 
//          configs given by m0 and m1 are in a safe position
// ------------------------------------------------------------
bool 
IVSimulator::areSafe( SoPathList& jks0,const Milestone& m0,
		      SoPathList& jks1,const Milestone& m1)
{
  // Moves the back robots to the position
  setConfig(jks0,m0.getConfig()); 
  setConfig(jks1,m1.getConfig());

  int i, j;
  // For jks0, get all of their links
  for(i=0;i<jks0.getLength();i++){
    JointKit *jk0=(JointKit *)(((SoNodeKitPath *)jks0[i])\
			       ->getNodeFromTail(0));
    Link *link0 = (Link *)jk0->getPart("link", FALSE);

    // For jks1, get all of their links
    for(j=0;j<jks1.getLength();j++){
      JointKit *jk1=(JointKit *)(((SoNodeKitPath *)jks1[j])\
				 ->getNodeFromTail(0));
      Link *link1 = (Link *)jk1->getPart("link", FALSE);
      // Check both links for collision
      if (!link0->isClear(link1))
	return false;
    }
  }
  return true;
}

bool 
IVSimulator::areSafe(const Milestone& m0, const Milestone& m1)
{
  // Moves the back robots to the position
  setConfig(brJointKits[0],m0.getConfig()); 
  setConfig(brJointKits[1],m1.getConfig());

  int i, j;
  // For jks0, get all of their links
  for(i=0;i<brJointKits[0].getLength();i++){
    JointKit *jk0=(JointKit *)(((SoNodeKitPath *)brJointKits[0][i])\
			       ->getNodeFromTail(0));
    Link *link0 = (Link *)jk0->getPart("link", FALSE);

    // For jks1, get all of their links
    for(j=0;j<brJointKits[1].getLength();j++){
      JointKit *jk1=(JointKit *)(((SoNodeKitPath *)brJointKits[1][j])\
				 ->getNodeFromTail(0));
      Link *link1 = (Link *)jk1->getPart("link", FALSE);
      // Check both links for collision
      if (!link0->isClear(link1))
	return false;
    }
  }
  return true;
}
// -----------------------------------------------------------------
// GetConfig: Reads the current config of the robot given by 'robID'
// -----------------------------------------------------------------
void 
IVSimulator::GetConfig(GenCoords *q, int robID) {
  int i, j;
  for (i = 0, j = 0; i < jointKitsR[robID].getLength(); i++) {
    JointKit *jk = (JointKit *)(((SoNodeKitPath *)jointKitsR[robID][i])->getNodeFromTail(0));
    jk->updateMatrix(true);
    Joint *joint = (Joint *)jk->getPart("joint", FALSE);
    if (joint) {
      joint->getVals(q, j);
    }
  }
  
  for (i = 0, j = 0; i <jointKitsR[robID].getLength(); i++) {
    JointKit *jk = (JointKit *)(((SoNodeKitPath *)jointKitsR[robID][i])->getNodeFromTail(0));
    jk->updateMatrix(true);};
}

// ------------------------------------------------------------
// SetConfig: Used to fix the configuration of robot 'robID'
// ------------------------------------------------------------
void 
IVSimulator::SetConfig(const GenCoords& q, int robID) {
  setConfig(jointKitsR[robID], q);
}

// ------------------------------------------------------------
// getCurDistance: Computes the distance from robot robotID
//                 to the closest object in Workspace
// ------------------------------------------------------------
float 
IVSimulator::getCurDist(int robID) {

  float minDist = 10e10;
  
  for(int j =0; j<objs;j++){
    for (int i = 0; i < jointKitsR[robID].getLength(); i++) {
      JointKit *jk = (JointKit *)(((SoNodeKitPath *)jointKitsR[robID][i])
				  ->getNodeFromTail(0));
      Link *link = (Link *)jk->getPart("link", FALSE);
      if (link) {
	float newDist = link->getCurDist(shapeModObst[j],0.0);
	if ( newDist<=0) 
	  return 0;
	if (newDist < minDist)
	  minDist = newDist;
      }
    }
       cerr<<"Distance to obstacle "<<j<<" : " << minDist<<endl;
  }
  return minDist;
}


// ------------------------------------------------------------------
// The separator read from the file is composed of a set of
// ObstacleKit, each of them contains a transformation node and an
// object from the class Obstacle, which is composed 
// by the Collision-Checking model and the actual shape or 
// geometry. This function takes apart all the parts
// in global variables in order to avoid having to search on the 
// scene path each time.
// -------------------------------------------------------------------
void IVSimulator::
decomposeObjKit(SoPathList obKits)
{

 SbVec3f t ;
  for (int i = 0; i < obKits.getLength(); i++) {
    ObstacleKit *jk = (ObstacleKit *)(((SoNodeKitPath *)obKits[i])->
                                      getNodeFromTail(0));
    obstacleSep[objs] = (SoSeparator *)(((SoNodeKitPath *)obKits[i])->
                                      getNodeFromTail(0));      
    root->addChild(obstacleSep[objs]);
 
    // Get the transformation for the shape
    xformObst[objs] = (SoTransform *)jk->getPart("transform", FALSE);
    if (xformObst[objs]){
      // The initial position from the iv file
      t = xformObst[objs]->translation.getValue();
    }
 
    // Get the Obstacle (col Model + geometric shape)
    shapeModObst[objs]= (Obstacle *)jk->getPart("shape", FALSE);

    if (shapeModObst[objs]){
      
      // Get the collision model
      shapeModObst[objs]->buildColModel();
      
      // sincronize its value with the one read from the iv file
      shapeModObst[objs]->setTranslation(t[0],t[1],t[2]);
    }
    objs++;
  }                
}

// --------------------------------------------
// --------------------------------------------
const char *
IVSimulator::getJointKitName(int jidx) 
{
  if ((jidx >= 0) && (jidx < jointKitsR[0].getLength())) {
    JointKit *kit = 
      (JointKit *)(((SoNodeKitPath *)jointKitsR[0][jidx])->getNodeFromTail(0));
    return kit->getName().getString();
  }
  return NULL;
}
// -----------------------------------------------------
// printConfig: Used to print the config given by q
// -----------------------------------------------------
void 
IVSimulator::printConfig(const GenCoords& q) {
  for (int i = 0; i < q.Dim(); i++)
    printf("%.4f\t",q[i]);
  printf("\n");
}

// ----------------------------------------------------------------
// required by inventor to initialize all nodes created by the user
// ----------------------------------------------------------------
void 
IVSimulator::initInventor() {
  JointKit::initClass();
  Joint::initClass();
  JointEngine::initClass();
  JointInvEngine::initClass();
  NullJoint::initClass();
  PlanarJoint::initClass();
  PlanarJointEngine::initClass();
  PlaneOrientJoint::initClass();
  PlaneOrientJointEngine::initClass();
  PrisJoint::initClass();
  PrisJointEngine::initClass();
  RevJoint::initClass();
  RevJointEngine::initClass();
  UnivJoint::initClass();
  UnivJointEngine::initClass();
  UnivJointInvEngine::initClass();
  Link::initClass();
  ObstacleKit::initClass();
  Obstacle::initClass();
}

// ----------------------------------------------------------------
// Once the file with the info of the robot is read, we have to 
// find the jointkits that compose the robot. A jointkit is specified
// by its geometric shape, limits of the joint, velocity, etc.
// ----------------------------------------------------------------
void 
IVSimulator::findJointKits(SoNode *top, SoPathList& jointKits) {
  SoSearchAction jkSearchAction; 
  jkSearchAction.setType(JointKit::getClassTypeId()); 
  jkSearchAction.setInterest(SoSearchAction::ALL);
  jkSearchAction.setSearchingAll(TRUE);
  jkSearchAction.apply(top);
	
  jointKits = jkSearchAction.getPaths();
}

// -----------------------------------------------------------------
// This is used with the same purpose than findJointKits, but in this
// case, for the objects. The objects (obstacles) are composed of its
// geometric shape and a transformation that defines its position
// -----------------------------------------------------------------
void 
IVSimulator::findObstacleKits(SoNode *top, SoPathList& obstKit) {
  SoSearchAction obstSearchAction; // action for search
  obstSearchAction.setType(ObstacleKit::getClassTypeId());// look for
  obstSearchAction.setInterest(SoSearchAction::ALL);// take the first
  obstSearchAction.setSearchingAll(TRUE);//search using normal traversal
  obstSearchAction.apply(top);
  obstKit = obstSearchAction.getPaths();
}

// -----------------------------------------------------------------
// Used to define a configuration for the robot given by jks
// -----------------------------------------------------------------
void 
IVSimulator::setConfig(SoPathList& jks, const GenCoords& q) {
  int i, j;
  for (i = 0, j = 0; i < jks.getLength(); i++) {
    JointKit *jk = 
      (JointKit *)(((SoNodeKitPath *)jks[i])->getNodeFromTail(0));
    Joint *joint = (Joint *)jk->getPart("joint", FALSE);
    joint->setVals(&q, j);
    jk->updateMatrix(false);
  
  }

  if (jks.getLength() > 0) {
    JointKit *jk = 
      (JointKit *)(((SoNodeKitPath *)jks[0])->getNodeFromTail(0));
    jk->updateMatrix(true);
  }
}

// -----------------------------------------------------------------
// Used to specify the position of an object 
// -----------------------------------------------------------------
void 
IVSimulator::setLocation(int obst_Nr, double x, double y, double z){
    // Update the position for the shape
    shapeModObst[obst_Nr]->setTranslation(x,y,z);
    // Update the position for the collision Model
    xformObst[obst_Nr]->translation.setValue(x,y,z);
}

// -----------------------------------------------------------------
// Used to fix the root of the scene graph
// -----------------------------------------------------------------
void 
IVSimulator::setRoot(SoSeparator *root_) {
  root = root_;

  // Add the robot Separators
  for (int k=0;k<ROBOTS;k++)
    root->addChild(robotSep[k]);
  
  // Add the Objects
  for(int j=0;j<MAX_OBJS;j++){
    if(obstacleSep[j])
      root->addChild(obstacleSep[j]);
  }
}

// ----------------------------------------------------------------
// Reads values as the velocity for each link, and the limits that
// define the kinematic model of the robot
// ----------------------------------------------------------------
void
IVSimulator::initialize()
{
  GenCoords min, max;

  // cache the max joint velocities
  int vj=0, minj=0, maxj=0, i;
  for (int k =0;k<ROBOTS;k++){  
    for (i=0; i < brJointKits[k].getLength(); i++) {
      JointKit *jk = (JointKit *) (((SoNodeKitPath *)brJointKits[k][i])\
				   ->getNodeFromTail(0));
      Joint *joint = (Joint *)jk->getPart("joint", FALSE);
      if (joint) {
	joint->getVmaxs(&maxVelocity, vj);
	joint->getMins(&min, minj);
	joint->getMaxs(&max, maxj);
      }
      // cache the joint interval
      jointInterval = max - min;
    }
  }
}

// ----------------------------------------------------------------
// All the configurations are composed as a set of n-tuplas where
// each element has a value >=0 and <=1. This function normalizes
// the values for each joint
// ----------------------------------------------------------------
void
IVSimulator::normalizeConfig(const GenCoords& q, GenCoords& q0)
{
  for (int i = 0, j = 0, k = 0; i < brJointKits[0].getLength(); i++) {
    JointKit *jk = (JointKit *)
      (((SoNodeKitPath *)brJointKits[0][i])->getNodeFromTail(0));                
    Joint *joint = (Joint *)jk->getPart("joint", FALSE);
    if (joint) {
      joint->setActVals(q.x, j);
      joint->getVals(&q0, k);
    }
  }
}

// ------------------------------------------------------------------
// To convert from the normalized value for the angle on each joint to
// the real value in degrees for the robot
// ------------------------------------------------------------------
void
IVSimulator::denormalizeConfig(const GenCoords& q0, GenCoords& q)
{
  for (int i = 0, j = 0, k = 0; i < brJointKits[0].getLength(); i++) {
    JointKit *jk = (JointKit *)
      (((SoNodeKitPath *)brJointKits[0][i])->getNodeFromTail(0));                
    Joint *joint = (Joint *)jk->getPart("joint", FALSE);
    if (joint) {
      joint->setVals(&q0, j);
      joint->getActVals(q.x, k);
    }
  }
}













