#include <stdio.h>
#include <string.h>
#include <math.h>
#include "pumaivsimulator.h"
#include "puma.h"

double PumaIVSimulator::pmscale[6] = {
	-0.10329,
	0.05978,
	-0.11060,
	0.41405,
	0.47696,
	0.46387,
};

/* the max toque is 10/mscale */

double PumaIVSimulator::pmaxtq[6] = {
	96.81479,	
	167.28003,
	90.41591,
	24.15167,
	20.96612,
	21.5577,
};


/* bases of the three pumas */
Frame3	PumaIVSimulator::pbase[3] = {
	{
		{{1.0, 0.0, 0.0},
		 {0.0, 1.0, 0.0},
		 {0.0, 0.0, 1.0}},
		{0.0, 0.0, 0.0}
	},
	{
		{{-1.0, 0.0, 0.0},
		 {0.0, -1.0, 0.0},
		 {0.0, 0.0, 1.0}},
		{827.1, 1343.2, 0.0}
	},
	{
		{{1.0, 0.0, 0.0},
		 {0.0, 1.0, 0.0},
		 {0.0, 0.0, 1.0}},
		{1787, 762, 0.0}
	},
};

/*
 * D&H paramters as per Armstrong et al but in mm
 */
#define D0	660.4	/* height of puma */
#define D2	243.5
#define	A2	431.8
#define D3	(-93.4)
#define A3	(-20.32)
#define D4	433.07
/* this is sqrt(A3*A3 + D4*D4) */
#define L3	433.54645

void
PumaIVSimulator::pkin(Frame3 f[7], const Frame3 *b, const double q[6])
{
	double c,s;

	memset(f, 0, 7*sizeof(Frame3));

	f[0] = *b;
	f[0].p.z += D0;

	c = cos(q[0]);
	s = sin(q[0]);

	f[1].r[0][0] = c;
	f[1].r[0][1] = -s;
	f[1].r[1][0] = s;
	f[1].r[1][1] = c;
	f[1].r[2][2] = 1;

	fmul3(&f[1], &f[0], &f[1]);

	c = cos(q[1]);
	s = sin(q[1]);

	f[2].r[0][0] = c;
	f[2].r[0][1] = -s;
	f[2].r[1][2] = 1;
	f[2].r[2][0] = -s;
	f[2].r[2][1] = -c;
	f[2].p.y = D2;

	fmul3(&f[2], &f[1], &f[2]);

	c = cos(q[2]);
	s = sin(q[2]);

	f[3].r[0][0] = c;
	f[3].r[0][1] = -s;
	f[3].r[1][0] = s;
	f[3].r[1][1] = c;
	f[3].r[2][2] = 1;
	f[3].p.x = A2;
	f[3].p.z = D3;

	fmul3(&f[3], &f[2], &f[3]);

	c = cos(q[3]);
	s = sin(q[3]);

	f[4].r[0][0] = c;
	f[4].r[0][1] = -s;
	f[4].r[1][2] = -1;
	f[4].r[2][0] = s;
	f[4].r[2][1] = c;
	f[4].p.x = A3;
	f[4].p.y = -D4;

	fmul3(&f[4], &f[3], &f[4]);

	c = cos(q[4]);
	s = sin(q[4]);

	f[5].r[0][0] = c;
	f[5].r[0][1] = -s;
	f[5].r[1][2] = 1;
	f[5].r[2][0] = -s;
	f[5].r[2][1] = -c;
	
	fmul3(&f[5], &f[4], &f[5]);

	c = cos(q[5]);
	s = sin(q[5]);
	
	f[6].r[0][0] = c;
	f[6].r[0][1] = -s;
	f[6].r[1][2] = -1;
	f[6].r[2][0] = s;
	f[6].r[2][1] = c;

	fmul3(&f[6], &f[5], &f[6]);
}

int	
PumaIVSimulator::pinvkin(double q[6], const Frame3 *f, int mode)
{
	double t, R2, R, r, s[6], c[6], s12, c12, a, b;
	Point3 p, p2;

	p = f->p;
	p.z -= D0;
	p2.x = p.x*p.x;	
	p2.y = p.y*p.y;
	p2.z = p.z*p.z;

	/* joint 0 */

	r = p2.x + p2.y - (D2+D3)*(D2+D3);
	/* printf("r = %f\n", r); */
	if (r < 0)
		return 0;
	r = sqrtf(r);
	if (mode & PRIGHT)
		r = -r;
	s[0] = (p.y*r - p.x*(D2+D3))/(p2.x+p2.y);
	c[0] = (p.x*r + p.y*(D2+D3))/(p2.x+p2.y);

	q[0] = atan2(s[0], c[0]);


	/* joint 1 */


	R2 = p2.x + p2.y + p2.z  - (D2+D3)*(D2+D3);
/* printf("R = %f [%f,%f]\n", sqrtf(R2), (L3-A2), (A2+L3)); */

	if (R2 < (L3-A2)*(L3-A2) || R2 > (A2+L3)*(A2+L3))
		return 0;
	R = sqrtf(R2);
	
	a = atan2(-p.z, r);
	b = acos((A2*A2 + R2 - L3*L3)/(2*A2*R));
	if (!(mode & PRIGHT) ^ !(mode & PEUP))
		b = -b;
	q[1] = a+b;
	if (q[1] < -MYPI*3/2)
		q[1] += 2*MYPI;
	else if (q[1] > MYPI/2)
		q[1] -= 2*MYPI;


	/* joint 2 */


	a = (R2 - A2*A2 - L3*L3)/(2*A2*L3);
	t = sqrtf(1-a*a);
	if (!(mode & PRIGHT) ^ !(mode & PEUP)) {
		t = -t;
	}
	q[2] =  -atan2(A3,D4)+atan2(a,t);
	if (q[2] < -MYPI/2)
		q[2] += 2*MYPI;
	else if (q[2] > MYPI*3/2)
		q[2] -= 2*MYPI;

	s[2] = sin(q[2]);
	c[2] = cos(q[2]);


	/* joint 3 */


	s12 = sin(q[1]+q[2]);
	c12 = cos(q[1]+q[2]);
	
	s[3] = -f->r[0][2]*s[0]+f->r[1][2]*c[0];
	c[3] = f->r[0][2]*c[0]*c12+f->r[1][2]*s[0]*c12-f->r[2][2]*s12;

	if (c[3] != 0 && s[3] != 0)
		q[3] = atan2(s[3],c[3]);
	else
		q[3] = 0;

	s[3] = sin(q[3]);
	c[3] = cos(q[3]);

	/* joint 4 */

	s[4] = f->r[0][2]*(c[0]*c12*c[3] - s[0]*s[3]) +
	       f->r[1][2]*(s[0]*c12*c[3] + c[0]*s[3]) +
	       -f->r[2][2]*c[3]*s12;
	c[4] = f->r[0][2]*c[0]*s12 +
	       f->r[1][2]*s[0]*s12 +
	       f->r[2][2]*c12;

	q[4] = atan2(s[4], c[4]);

	
	/* joint 5 */
		

	s[5] = f->r[0][0]*(-s[0]*c[3]-c[0]*c12*s[3]) +
	       f->r[1][0]*(c[0]*c[3]-s[0]*c12*s[3]) +
	       f->r[2][0]*s[3]*s12;
	c[5] = f->r[0][1]*(-s[0]*c[3]-c[0]*c12*s[3]) +
	       f->r[1][1]*(c[0]*c[3]-s[0]*c12*s[3]) +
	       f->r[2][1]*s[3]*s12;
	q[5] = atan2(s[5], c[5]);


	/* figure out whether to flip the wrist */

	if (mode & PWDOWN) {
		if (q[4] > 0.0) {
			q[4] = -q[4];
			if (q[3] > 0)
				q[3] -= MYPI;
			else
				q[3] += MYPI;
			if (q[5] > 0)
				q[5] -= MYPI;
			else
				q[5] += MYPI;
		}
	} else if (q[4] < 0.0) {
		q[4] = -q[4];
		if (q[3] > 0)
			q[3] -= MYPI;
		else
			q[3] += MYPI;
		if (q[5] > 0)
			q[5] -= MYPI;
		else
			q[5] += MYPI;
	}

	return 1;
}

void
PumaIVSimulator::pendeff(Frame3 *f, double q[6])
{
/*  This function calculates the position of the
 *  wrist intersection point for the PUMA 560 based 
 *  on the current joint angles.
 */

	double s12, c12;
	double R;
	double to1,to2,to3,to4;
	double alpha1, alpha2;
	double beta1, beta2;
	double gamma1, gamma2; 
	double delta;
	double eta1, eta2;
	double s[6], c[6];
	int i;
	
	for (i = 0; i < 6; i++) {
		c[i] = cos(q[i]);
		s[i] = sin(q[i]);
	}
 
	s12 = s[1]*c[2] + s[2]*c[1];
	c12 = c[1]*c[2] - s[1]*s[2];

	/* Calculate position of wrist point */

	R   = D4*s12 + A3*c12 + A2*c[1];

	f->p.x = c[0]*R - s[0]*(D2+D3);
	f->p.y = s[0]*R + c[0]*(D2+D3);
	f->p.z = D4*c12 - A3*s12 - A2*s[1] + D0;

	/* Calculate orientation of wrist */

	to1 = c12*s[3];
	to2 = s12*c[3];
	to3 = c12*c[3];
	to4 = s[3]*c[4];
	eta2 = s12*s[3];
	gamma2 = s[3]*s[4];
	eta1 = to2*c[4] + c12*s[4];
	gamma1 = to3*s[4] + s12*c[4];
	delta = to3*c[4] - s12*s[4];
	alpha1 = delta*c[5] - to1*s[5];
	alpha2 = c[3]*s[5] + to4*c[5];
	beta1 = delta*s[5] + to1*c[5];
	beta2 = c[3]*c[5] - to4*s[5];

	f->r[0][0] = c[0]*alpha1 - s[0]*alpha2;
	f->r[1][0] = s[0]*alpha1 + c[0]*alpha2;
	f->r[2][0] = -eta1*c[5] + eta2*s[5];

	f->r[0][1] = -c[0]*beta1 - s[0]*beta2;
	f->r[1][1] = -s[0]*beta1 + c[0]*beta2;
	f->r[2][1] = eta1*s[5] + eta2*c[5];

	f->r[0][2] = c[0]*gamma1 - s[0]*gamma2;
	f->r[1][2] = s[0]*gamma1 + c[0]*gamma2;
	f->r[2][2] = -to2*s[4] + c[4]*c12;
}

void
PumaIVSimulator::pjacobian(double J[6][6], double q[6])
{
/*  This function calculates the jacobian of the
 *  wrist intersection point for the PUMA 560 based 
 *  on the current joint angles.
 */

	double s12, c12;
	double R, R12, Rz;
	double to1,to2,to3,to4;
	double alpha1, alpha2;
	double beta1, beta2;
	double gamma1, gamma2; 
	double delta;
	double eta1, eta2;
	double s[6], c[6];
	int i;
	
	for (i = 0; i < 6; i++) {
		c[i] = cos(q[i]);
		s[i] = sin(q[i]);
	}
 
	s12 = s[1]*c[2] + s[2]*c[1];
	c12 = c[1]*c[2] - s[1]*s[2];

	R12 = D4*s12 + A3*c12;
	Rz  = D4*c12 - A3*s12;
	R   = R12 + A2*c[1];

	to1 = c12*s[3];
	to2 = s12*c[3];
	to3 = c12*c[3];
	to4 = s[3]*c[4];
	eta2 = s12*s[3];
	gamma2 = s[3]*s[4];
	eta1 = to2*c[4] + c12*s[4];
	gamma1 = to3*s[4] + s12*c[4];
	delta = to3*c[4] - s12*s[4];
	alpha1 = delta*c[5] - to1*s[5];
	alpha2 = c[3]*s[5] + to4*c[5];
	beta1 = delta*s[5] + to1*c[5];
	beta2 = c[3]*c[5] - to4*s[5];

	J[0][0] = -s[0]*R - c[0]*(D2+D3);
	J[0][1] = (D4*c12 - A3*s12 - A2*s[1])*c[0];
	J[0][2] = c[0]*Rz;
	J[0][3] = 0.0;
	J[0][4] = 0.0;
	J[0][5] = 0.0;

	J[1][0] = c[0]*R - s[0]*(D2+D3);
	J[1][1] = (D4*c12 - A3*s12 - A2*s[1])*s[0];
	J[1][2] = s[0]*Rz;
	J[1][3] = 0.0;
	J[1][4] = 0.0;
	J[1][5] = 0.0;

	J[2][0] = 0.0;
	J[2][1] = -R;
	J[2][2] = -R12;
	J[2][3] = 0.0;
	J[2][4] = 0.0;
	J[2][5] = 0.0;

	J[3][0] = 0.0;
	J[3][1] = -s[0];
	J[3][2] = -s[0];
	J[3][3] = c[0]*s12;
	J[3][4] = -c[0]*c12*s[3] - s[0]*c[3];
	J[3][5] = c[0]*gamma1 - s[0]*s[3]*s[4];

	J[4][0] = 0.0;
	J[4][1] = c[0];
	J[4][2] = c[0];
	J[4][3] = s[0]*s12;
	J[4][4] = -s[0]*c12*s[3] + c[0]*c[3];
	J[4][5] = s[0]*gamma1 + c[0]*s[3]*s[4];;

	J[5][0] = 1.0;
	J[5][1] = 0.0;
	J[5][2] = 0.0;
	J[5][3] = c12;
	J[5][4] = s12*s[3];
	J[5][5] = -s12*c[3]*s[4] + c[4]*c12;
}


