/* ********************************************************************
   File pumaivsimulator.C
   The functions implemented in this file define the
   Forward and Inverse Kinematics for a PUMA Robot

   Modified by Gildardo Sanchez-Ante
   Robotics Lab., Stanford University
   ante@robotics.stanford.edu
   *******************************************************************/

#include <iostream.h>
#include <math.h>

#include "pumaivsimulator.h"
#include "puma.h"
#include "joint.h"
#include "jointkit.h"
#include "Path.h"

PumaIVSimulator::PumaIVSimulator(SoSeparator *root):
  BaseIVSimulator(root) 
{
  GenCoords::Dim(6);

  fromIV = iframe3[0];
  fromIV.r[1][1] = 0;
  fromIV.r[2][2] = 0;
  fromIV.r[1][2] = 1;
  fromIV.r[2][1] = 1;
  finv3(&toIV, &fromIV);
  baseFrame = iframe3[0];
}

PumaIVSimulator::~PumaIVSimulator() {}

bool PumaIVSimulator::IsClear(const Milestone& m)
{
  return IVSimulator::IsClear(m);
}

bool 
PumaIVSimulator::IsClear(const Milestone& m0, const Milestone& m1, 
			 IntervalType itype) 
{  
  //cerr<<"IN PumaIVSimulator::IsClear"<<endl;
  // compute  the discretization for each joint of the robot
  int numSteps=1, n;

  GenCoords q0(m0.getConfig()),q1(m1.getConfig());
  
  int i;
  for (i=0; i<3; i++) {		                  // first three joints
     n = fabs( (q1[i]-q0[i])*jointInterval[i] )*15+1; // check every 4 deg
     numSteps = numSteps > n ? numSteps : n;
  }

  for (i=3; i<6; i++) {			                 // last three joints
    n = fabs( (q1[i]-q0[i])*jointInterval[i] )*4+1;  // check every 15 deg
    numSteps = numSteps > n ? numSteps : n;
  }

  int start, end;

  switch (itype) {
  case Open: {
    start = 1;
    end = numSteps-1;
    break;
  }
  case SemiOpen: {

    start = 1;
    end = numSteps;
    break;
  }
  case SemiClosed: {
    start = 0;
    end = numSteps-1;
    break;
  }
  case Closed:
    
  default: {
    start = 0;
    end = numSteps;
    break;
  }
  }
  
  //cerr<<numSteps<<endl;

  GenCoords q; double t,t0,t1;
  t0=m0.getTime(); t1=m1.getTime();

  for (i = start; i <= end; i++) {
    q = q0+((double)i/numSteps)*(q1-q0);
    t = t0+(i*(t1-t0));

    Milestone m(0,q,t);
    if (!IsClear(m))
      return false;
  }
  
  return true;
    
}


inline double
PumaIVSimulator::Distance(const GenCoords& p, const GenCoords& q)
{
  int i=GenCoords::Dim();		// number of dofs
  double l=0;
  
  while (i--) 
    l += (p[i]-q[i])*(p[i]-q[i]);

  return sqrt(l);
}


bool PumaIVSimulator::IK(const Pose& x, GenCoords *q_, const GenCoords *q0) 
{
  Frame3 f, xf;
  poseToFrame(xf, x);
  fmul3(&f, &ibaseFrame, &xf);
  fmul3(&f, &f, &toIV);
  fmul3(&f, &fromIV, &f);

  const int NUM_IK_SOLS = 8;
  GenCoords qq[NUM_IK_SOLS], qtmp;
  double    dmin = 100;
  int       qidx = -1;

  // find the solution closest to q0
  for (int m=0; m < NUM_IK_SOLS; m++) {
      
    if ( pinvkin(qtmp.x, &f, m) ) {
      normalizeConfig(qtmp, qq[m]);
      double d = Distance( qq[m], *q0 )  ;
      if (d < dmin) {
	qidx = m;
	dmin = d;
      }
    }
  }

  if (qidx == -1) 
    return false;
  else {
    *q_ = qq[qidx];
    return true;
  }
}

bool PumaIVSimulator::IK(const Pose& x, GenCoords *q_, int mode) 
{
  Frame3 f, xf;
  poseToFrame(xf, x);
  fmul3(&f, &ibaseFrame, &xf);
  fmul3(&f, &f, &toIV);
  fmul3(&f, &fromIV, &f);

  GenCoords  qtmp;
  if  ( pinvkin(qtmp.x, &f, mode) ) {
    normalizeConfig(qtmp, *q_);
    return true;
  }
  else
    return false;
}

void PumaIVSimulator::FK(const GenCoords& q, Pose *x, int rob) {
  //	Frame3 f[7], b;
  //	double q[6];

  setConfig(brJointKits[rob], q);
 
  /*
    for (int i = 0, j = 0; i < brJointKits.getLength(); i++) {
    JointKit *jk = (JointKit *)(((SoNodeKitPath *)brJointKits[i])->getNodeFromTail(0));
    jk->updateMatrix(false);
    Joint *joint = (Joint *)jk->getPart("joint", FALSE);
    if (joint) {
    joint->getActVals(q, j);
    }
    }
    
    b =iframe3[0];
    pkin(f, &b, q);
    fmul3(&f[6], &f[6], &toIV);
    fmul3(&f[6], &fromIV, &f[6]);
    fmul3(&f[6], &baseFrame, &f[6]);
    frameToPose(*x, f[6]);
  */
  JointKit *jk = (JointKit *)
    (((SoNodeKitPath *)brJointKits[rob][6])->getNodeFromTail(0));
  x->setValue(jk->getAccuMatrix().getValue());
}

void PumaIVSimulator::SetBase(const Pose& basePose_) {
	BaseIVSimulator::SetBase(basePose_);
	// For all the robots
	for(int i=0;i<10;i++){
	  poseToFrame(baseFrame, basePose[i]);
	  poseToFrame(ibaseFrame, basePose[i]);
	  finv3(&ibaseFrame, &ibaseFrame);
	}
}

void PumaIVSimulator::poseToFrame(Frame3 &frame, const Pose& pose) {
	for (int i = 0; i < 3; i++) { 
		for (int j = 0; j < 3; j++) {
			frame.r[j][i] = pose[i][j];
		}
	}

	frame.p.x = pose[3][0]*1000;
	frame.p.y = pose[3][1]*1000;
	frame.p.z = pose[3][2]*1000;
}

void PumaIVSimulator::frameToPose(Pose& pose, const Frame3& frame) {
	pose.makeIdentity();
	for (int i = 0; i < 3; i++) { 
	  for (int j = 0; j < 3; j++) {
	    pose[i][j] = frame.r[j][i];
	  }
	}

	pose[3][0] = frame.p.x/1000.;
	pose[3][1] = frame.p.y/1000.;
	pose[3][2] = frame.p.z/1000.;
}

double
PumaIVSimulator::Cost(const GenCoords& p, const GenCoords& q)
{
  double c=0.0, ci;
  for (int i=0; i < GenCoords::Dim(); i++) {
    ci = (q[i]-p[i]) / maxVelocity[i];
    ci = ci > 0.0 ? ci : -ci;
    c  = ci > c ? ci : c;
  }

  return c;
}


void
PumaIVSimulator::InterpolatePath(const Path& p, Path *ip, double speed)
{
  double dt  = 1.0/30*speed;            // time in between frames
  double ttotal = 0.0;
  int i=0;

  double t0=0;
  Path::const_iterator s0, s1;
  s0=s1=p.begin();
  for (s1++; s1 != p.end(); s0++, s1++) {

    GenCoords ds = (*s1).getConfig() - (*s0).getConfig();
    GenCoords ti= ds / maxVelocity;
    double tmax = ti.Max();
    GenCoords vel = ds / tmax;
    i=0;
    double t;
    for (t=t0; t < tmax; t += dt) {
 
      GenCoords q=(*s0).getConfig() + t * vel; 
      double t1 = (*s0).getTime()+i*dt;
     
      Milestone m(i,q,t1);
      ip->push_back(m);
      i++;
    }
    
    t0 = t - tmax;
    ttotal += tmax;
  }

  ip->push_back( p.back() );            // insert the last config

  //  cerr << "Total simulation time = " << ttotal << endl;
}


double
PumaIVSimulator::WspDist(const GenCoords& p, const GenCoords& q)
{
  double c=0.0;

  for (int i=0; i<3; i++) {
    //   double t= (q[i]-p[i])*jointInterval[i];
   double t= (q[i]-p[i]);
    c += t*t;
  }

  return sqrt(c);
}


double
PumaIVSimulator::NeighDist2(const GenCoords& p, const GenCoords& q)
{
  double c=0.0;

  for (int i=0; i<6; i++) {
    //    double t= (q[i]-p[i])*jointInterval[i];
    double t= (q[i]-p[i]);
    c += t*t;
  }

  return c;
}







